MODULE SIO; (*07.11.94. LB*)
(*Last modification: 7. September 1995, LB*)

(* Implements the procedures to read and write streams.
   Error handling is somewhat strange: the idea was 
   to protect beginners from being confronted with exceptions. *)
  IMPORT
    Rd, Wr, Stdio, Lex, Fmt, Thread, FloatMode, Text;

<*FATAL Thread.Alerted, Wr.Failure*>

  VAR
    termChar: CHAR := null; lastReader: Reader := NIL;
  TYPE
    Kind = {Integer, Real, LongReal, Boolean};

  PROCEDURE LookAhead(rd: Reader := NIL): CHAR RAISES {Error} =
  (*Returns the next character, without removing it from the reader.*)
  VAR ch: CHAR := null;
  BEGIN
    IF rd = NIL THEN rd:= Stdio.stdin END;
    TRY
      ch:= Rd.GetChar(rd);
      Rd.UnGetChar(rd);
    EXCEPT
      Rd.EndOfFile, Rd.Failure => RAISE Error;
    END; (*TRY*)
    RETURN ch;
  END LookAhead;

  PROCEDURE GetChar(rd: Reader := NIL): CHAR RAISES {Error} =
  (*Returns the next char.*)
  VAR ch: CHAR := null;
  BEGIN
    IF rd = NIL THEN rd:= Stdio.stdin END;
    TRY
      ch:= Rd.GetChar(rd);
    EXCEPT
      Rd.EndOfFile, Rd.Failure => RAISE Error;
    END; (*TRY*)
    RETURN ch;
  END GetChar;

  PROCEDURE PutChar(ch: CHAR; wr: Writer := NIL) =
  (*Outputs a single character.*)
  BEGIN
    IF wr = NIL THEN wr:= Stdio.stdout END;
    Wr.PutChar(wr, ch);
    IF wr = Stdio.stdout THEN Wr.Flush(wr) END;
  END PutChar;

  PROCEDURE ConsumeNext(rd: Reader) RAISES {Error} =
  (*Consumes the first terminating character after*)
  BEGIN
    IF NOT End(rd) THEN termChar:= GetChar(rd); lastReader:= rd END;
  END ConsumeNext;

  PROCEDURE GetText(rd: Reader := NIL): TEXT RAISES {Error} =
  (*Reads a text. The terminating character is not appended.*)
  VAR t: TEXT;
  BEGIN
    IF rd = NIL THEN rd:= Stdio.stdin END;
    TRY
      Lex.Skip(rd);
      t:= Lex.Scan(rd, Legibles);
    EXCEPT 
      Rd.Failure => RAISE Error
    END;
    ConsumeNext(rd);
    RETURN t
  END GetText;

  PROCEDURE PutText(t: TEXT; wr: Writer := NIL) =
  (*Outputs all characters in t.*)
  BEGIN
    IF wr = NIL THEN wr:= Stdio.stdout END;
    Wr.PutText(wr, t);
    IF wr = Stdio.stdout THEN Wr.Flush(wr) END;
  END PutText;

  PROCEDURE GetLine(rd: Reader := NIL): TEXT RAISES {Error} =
  (*Reads a line and returns is as a text.
    The terminating nl is not appended.*)
  VAR l: TEXT := "";
  BEGIN
    IF rd = NIL THEN rd:= Stdio.stdin END;
    TRY
      l:= Rd.GetLine(rd);
    EXCEPT 
      Rd.EndOfFile, Rd.Failure => RAISE Error;
    END; (*TRY*)
    termChar:= '\n'; lastReader:= rd;
    RETURN l;
  END GetLine;

  PROCEDURE PutLine(t: TEXT; wr: Writer := NIL) =
  (*Outputs all characters in t and appends a new line.*)
  BEGIN
    PutText(t & "\n", wr);
  END PutLine;

  PROCEDURE FormatError(rd: Reader; k: Kind;
                        VAR count: CARDINAL) RAISES {Error} =
  (*Handles badly formatted numerical and Boolean input*)
  CONST
    K = ARRAY Kind OF TEXT{"Integer", "Real", "LongReal", "Boolean"};
  VAR 
    T: TEXT := "\nThis is not a proper " & K[k] & ", try again\n";
  BEGIN
    IF rd = Stdio.stdin THEN
      REPEAT termChar:= GetChar(rd) UNTIL termChar = '\n';
      IF count > 0 THEN
        DEC(count);
        PutText(T);
      ELSE
        RAISE Error
      END;          
    ELSE
      Lex.Skip(rd, Legibles);
      RAISE Error 
    END; (*IF rd = Stdio.stdin*)
  END FormatError;

  PROCEDURE GetInt(rd: Reader := NIL): INTEGER RAISES {Error} =
  (*Reads a decimal integer.*)
  VAR i: INTEGER; count: CARDINAL := MaxError;
  BEGIN
    IF rd = NIL THEN rd:= Stdio.stdin END;
    LOOP
      TRY
        IF End(rd) THEN RAISE Error END;
        Lex.Skip(rd);
        i:= Lex.Int(rd);
        EXIT
      EXCEPT 
        Lex.Error, Rd.Failure, FloatMode.Trap =>
          FormatError(rd, Kind.Integer, count);
      END; (*TRY*)
    END; (*LOOP*)
    ConsumeNext(rd);
    RETURN i;
  END GetInt;
  
  PROCEDURE PutInt(i: INTEGER; length:= 3; wr: Writer := NIL) =
  (*Outputs an integer number. 
    The number is right-aligned in a field of length "length", i.e.
    leading blanks are output if the number of digits < "length".*)
  BEGIN
    IF wr = NIL THEN wr:= Stdio.stdout END;
    Wr.PutText(wr, Fmt.Pad(Fmt.Int(i), length));
    IF wr = Stdio.stdout THEN Wr.Flush(wr) END;
  END PutInt;

  PROCEDURE GetReal(rd: Reader := NIL): REAL RAISES {Error} =
  (*Reads a real number.*)
  VAR r: REAL; count: CARDINAL := MaxError;
  BEGIN
    IF rd = NIL THEN rd:= Stdio.stdin END;
    LOOP
      TRY
        IF End(rd) THEN RAISE Error END;
        Lex.Skip(rd);
        r:= Lex.Real(rd);
        EXIT;
      EXCEPT 
        Lex.Error, Rd.Failure, FloatMode.Trap => 
          FormatError(rd, Kind.Real, count);
      END; (*TRY*)
    END; (*LOOP*)
    ConsumeNext(rd);
    RETURN r;
  END GetReal;

  PROCEDURE PutReal(r: REAL; wr: Writer := NIL) =
  (*Outputs a real number. *)
  BEGIN
    IF wr = NIL THEN wr:= Stdio.stdout END;
    Wr.PutText(wr, Fmt.Real(r (*, style:= Fmt.Style.Sci *) ));
    (*Auto-mode doesn't work properly on the PC!*)
    IF wr = Stdio.stdout THEN Wr.Flush(wr) END;
  END PutReal;

  PROCEDURE GetLongReal(rd: Reader := NIL): LONGREAL RAISES {Error} =
  (*Reads a long real number.*)
  VAR lr: LONGREAL; count: CARDINAL := MaxError;
  BEGIN
    IF rd = NIL THEN rd:= Stdio.stdin END;
    LOOP
      TRY
        IF End(rd) THEN RAISE Error END; 
        Lex.Skip(rd);
        lr:= Lex.LongReal(rd);
        EXIT;
      EXCEPT 
        Lex.Error, Rd.Failure, FloatMode.Trap => 
          FormatError(rd, Kind.LongReal, count);
      END; (*TRY*)
    END; (*LOOP*)
    ConsumeNext(rd);
    RETURN lr;
  END GetLongReal;

  PROCEDURE PutLongReal(r: LONGREAL; wr: Writer := NIL) =
  (*Outputs a long real number. *)
  BEGIN
    IF wr = NIL THEN wr:= Stdio.stdout END;
    Wr.PutText(wr, Fmt.LongReal(r (*, style:= Fmt.Style.Sci*) ));
    (*Auto-mode doesn't work properly on the PC!*)
    IF wr = Stdio.stdout THEN Wr.Flush(wr) END;
  END PutLongReal;

  PROCEDURE Konv(t: TEXT): TEXT = 
  (*Converts lower case letters into capitals.*)
  CONST 
    Code   = ORD('A') - ORD('a');  
    Smalls = SET OF CHAR{'a' .. 'z'};
  VAR 
    t2: TEXT := "";   (*result*)
    ch: CHAR;
  BEGIN
    FOR i:= 0 TO Text.Length(t) - 1 DO
      ch:= Text.GetChar(t, i);
      IF ch IN Smalls THEN ch:= VAL(ORD(ch) + Code, CHAR) END;
      t2:= t2 & Text.FromChar(ch);
    END; (*FOR i*)
    RETURN t2
  END Konv;

  PROCEDURE GetBool(rd: Reader := NIL): BOOLEAN RAISES {Error} =
  (*Reads a Boolean constant, terminated by any character in "terminate".
   Legal values are: TRUE and FALSE and any shorthands of them.
   The case of letters is not significant.*)
  VAR 
    t: TEXT := ""; b: BOOLEAN;
    count: CARDINAL := MaxError;
  BEGIN
    IF rd = NIL THEN rd:= Stdio.stdin END;
    LOOP
      TRY
        IF End(rd) THEN RAISE Error END;
        Lex.Skip(rd);
        t:= Lex.Scan(rd, Legibles);
      EXCEPT 
        Rd.Failure => RAISE Error
      END;
      IF Text.Length(t) = 0 THEN t:= "x" ELSE t:= Konv(t) END;
      IF    Text.Equal(t, Text.Sub("TRUE",  0, Text.Length(t))) THEN
        b:= TRUE; EXIT
      ELSIF Text.Equal(t, Text.Sub("FALSE", 0, Text.Length(t))) THEN
        b:= FALSE; EXIT
      ELSE
        FormatError(rd, Kind.Boolean, count);
      END;
    END; (*LOOP*)
    ConsumeNext(rd);
    RETURN b
  END GetBool;
  
  PROCEDURE PutBool(b: BOOLEAN; wr: Writer := NIL) =
  (*Outputs a Boolean value.*)
  BEGIN
    IF wr = NIL THEN wr:= Stdio.stdout END;
    IF b THEN Wr.PutText(wr, "TRUE") ELSE Wr.PutText(wr, "FALSE") END;
    IF wr = Stdio.stdout THEN Wr.Flush(wr) END;
  END PutBool;

  PROCEDURE TermChar(rd: Reader := NIL): CHAR RAISES {Error} =
  (*Returns the last terminating character or null.
    At the end of the file TermChar returns the null character.
    There is a restriction on the usage of TermChar:
    it must be applied to the same reader as the last read operation,
    otherwise it returns the null character.
    Therefore, TermChar should not be used in a multi-threaded program.*)
  BEGIN
    IF rd = NIL THEN rd:= Stdio.stdin END;
    IF NOT End(rd) AND rd = lastReader THEN 
      RETURN termChar 
    ELSE 
      RETURN null 
    END;
  END TermChar;

  PROCEDURE Nl(wr: Writer := NIL) =
  (*Outputs a new line.*)
  BEGIN
    IF wr = NIL THEN wr:= Stdio.stdout END;
    Wr.PutText(wr, "\n");
    IF wr = Stdio.stdout THEN Wr.Flush(wr) END;
  END Nl;
  
  PROCEDURE PutUnsigned(i: INTEGER; length := 6; base: [2..16] := 16;
                        wr: Writer := NIL) =
  (*Outputs an unsigned number with given base.
    The number is right-aligned in a field of length "length".*)
  BEGIN
    IF wr = NIL THEN wr:= Stdio.stdout END;
    Wr.PutText(wr, Fmt.Pad(Fmt.Unsigned(i, base), length));
    IF wr = Stdio.stdout THEN Wr.Flush(wr) END;
  END PutUnsigned;

  PROCEDURE Flush(wr: Writer := NIL) =
  (*Flushes the writer on the file. Not necessary for Stdio.stdout.*)
  BEGIN
    IF wr = NIL THEN wr:= Stdio.stdout END;
    Wr.Flush(wr);
  END Flush;

  PROCEDURE End(rd: Reader := NIL): BOOLEAN =
  (*Returns TRUE iff end of reader reached.
    On the keyboard, CTRL-Z on the PC, and CTRL-D in Unix*)
  VAR eof: BOOLEAN := TRUE;
  BEGIN
    IF rd = NIL THEN rd:= Stdio.stdin END;
    TRY
      eof:= Rd.EOF(rd);
    EXCEPT ELSE
    END; (*TRY*)
    RETURN eof;
  END End; 

  PROCEDURE Available(rd: Reader := NIL): BOOLEAN =
  (*Returns TRUE iff some characters are available.*)
  VAR av: BOOLEAN := FALSE;
  BEGIN
    IF rd = NIL THEN rd:= Stdio.stdin END;
    TRY
      av:= Rd.CharsReady(rd) > 0
    EXCEPT ELSE
    END; (*TRY*)
    RETURN av;
  END Available;

  PROCEDURE Length(rd: Reader := NIL): CARDINAL =
  (*Returns the length of the reader.
    Returns 0 if the length cannot be computed.*)
  VAR l: CARDINAL := 0;
  BEGIN
    IF rd # NIL AND rd # Stdio.stdin THEN
      TRY
        l:= Rd.Length(rd)
      EXCEPT ELSE
      END; (*TRY*)
    END; (*IF*)
    RETURN l;
  END Length;

  PROCEDURE Seek(rd: Reader := NIL; position: CARDINAL := 0)
	 RAISES {Error} =
  (*Sets the reader on "position" if it is seekable.
    Default corresponds to reset a reader.*)
  BEGIN
    IF rd = NIL THEN rd:= Stdio.stdin END;
    IF Rd.Seekable(rd) THEN 
      TRY
        Rd.Seek(rd, position)
      EXCEPT ELSE
        RAISE Error
      END; (*TRY*)
    END; (*IF Rd.Seekable(rd)*)
  END Seek;

BEGIN
END SIO.


